
from Protocol import NullProto
from MsgEvent import MsgEvent
from Selecting.log import log

class Independent:
	def __init__(self):
		self.id = 0
		self.name = "Independent"

	def __str__(self):
		return "Independent"
	def __repr__(self):
		return "<Independent>"

	def send_event(self, event):
		pass
	def get_selfasevent(self, evtype):
		return MsgEvent(evtype, player=self, name=self.name)
	def think_about_death(self):
		pass # Independence never dies!
	def is_enemy(self, pl):
		return 1
	def bkeep_planet(self, planet):
		pass
	def ubkeep_planet(self, planet):
		pass

class Player:
	def __init__(self, server, id):
		self.server = server
		self.id = id
		self.game = None
		self.proto = NullProto(self)
		self.name = ""
		self.desc = ""
		self.alive = 1
		self.flying = []
		self.plns = []
		self.commands = {	'JOIN' : self.recv_join, \
					'MESG' : self.recv_message, \
					'MOVE' : self.recv_move, \
					'DONE' : self.recv_move, \
					'SUIC' : self.recv_suicide, \
					'QUIT' : self.recv_quit }

	def __str__(self):
		return "%d" % self.id

	def __repr__(self):
		return "<%d>" % self.id

	def regproto(self, proto):
		self.proto = proto

	def reg_game(self, game):
		if self.game != None:
			self.game.unreg_player(self)
		self.game = game
		self.proto.got_game(game)
		self.server.unreg_player_as_gamelistener(self)

	def unreg_game(self, game):
		self.game = None

	def is_alive(self, game):
		return self.alive

	def is_enemy(self, other):
		return self != other

	def is_attackable(self, other):
		return self.is_enemy(other)

	def send_event(self, event):
		log( 7, "%d OUP %s" % ( self.id, event ))
		self.proto.send_event(event)

	def get_selfasevent(self, evtype):
		return MsgEvent(evtype, name=self.name, player=self, desc=self.desc)

	def handshake_afterhook(self):
		self.server.list_games(self)

	def bkeep_move(self, ev):
		self.flying.append(ev)
	def ubkeep_move(self, ev):
		self.flying.remove(ev)
	def bkeep_planet(self, p):
		self.plns.append(p)
	def ubkeep_planet(self, p):
		self.plns.remove(p)

	def think_about_death(self):
		if len(self.flying) > 0:
			return
		for p in self.plns:
			if p.get_ships(self) > 0:
				return
		self.recv_event(MsgEvent('SUIC', player=self, game=self.game, desc="Rightful death"))

	# Here beginneth the command handler section
	def recv_event(self, event):
		log( 7, "%d INP %s" % ( self.id, event ))
		try:
			command = self.commands[event.evtype]
		except KeyError:
			if self.game == None:
				self.server.recv_event(event)
			else:
				self.game.recv_event(event)
			return
		command(event)

	def recv_move(self, event):
		if event.player != self:
			self.send_event(MsgEvent('ERRP', name=event.evtype, \
						 desc="Trying to make others' moves?"))
			return
		if not self.alive:
			self.send_event(MsgEvent('ERRP', name=event.evtype, \
						 desc="Don't move! You're dead."))
			return
		if self.game == None:
			self.send_event(MsgEvent('ERRP', name=event.evtype, \
						 desc="Please join a game first."))
			return
		self.game.recv_event(event)

	def recv_join(self, event):
		if event.player != self:
			self.send_event(MsgEvent('ERRP', name=event.evtype, \
						 desc="Join yourself."))
			return
		if self.game != None:
			self.send_event(MsgEvent('ERRP', name=event.evtype, \
						 desc="Multiple joining is not implemented."))
		self.name = event.name
		self.desc = event.desc
		event.game.recv_event(event)

	def recv_message(self, event):
		if event.player != self:
			self.send_event(MsgEvent('ERRP', name=event.evtype, \
						 desc="Look who's talking."))
			return

		if event.game != None:
			event.game.recv_event(event)
		else:
			self.server.recv_event(event)

	def recv_suicide(self, event):
		if event.player != self:
			self.send_event(MsgEvent('ERRP', name=event.evtype, \
						 desc="Killing someone else is not a suicide."))
			return
		self.alive = 0
		if self.game != None:
			self.game.recv_event(event)

	def recv_quit(self, event):
		self.alive = 0
		if self.game != None:
			self.game.recv_event(event)
			# May unregister player if game ends
		if self.game != None:
			self.game.unreg_player(self)
		if self.server != None:
			self.server.unreg_player(self.id)

